#!/usr/bin/perl -w

$confGmetric = "/usr/bin/gmetric --tmax=700";
$confGanglia = "/etc/gmond.conf";
$confDatafile = "/var/run/ganglia-vmstat.txt";
$confDatafile = "./ganglia-vmstat.txt" if ! -w "/var/run";

#die "Ganglia is not installed" if ! -r $confGanglia;

my $verbose = grep(/-v/,@ARGV);
my $dryrun  = grep(/-n/,@ARGV);

my $gmetricDest = "";

if (-r $confGanglia)
  {
    my $ip;
    my $port;
    
    open(IN,"$confGanglia") || die "Cannot read $confGanglia";
    while (<IN>)
      {
        next if /^#/;
        $ip   = $1 if /mcast_channel +(.*)/;
        $port = $1 if /mcast_port +(.*)/;
      }
    close IN;
    
    $gmetricDest .= " -c$ip " if defined $ip;
    $gmetricDest .= " -p$port " if defined $port;
  }

$ENV{"LANG"} = "C";
$ENV{"TERM"} = "vt100";

my $xtime = 0;

my $now = time();
my @vmstat = split(/\n/,`vmstat -s`);
my @diskstat = split(/\n/,`cat /proc/diskstats`);

foreach my $s (@vmstat)
  {
    next unless $s =~ /^\s+(\d+)\s+(\w+.+)$/;
    my $count = int($1);
    my $name  = $2;
    $name =~ s/\s/_/g;
    #print "[$count][$name]\n";
    $count{$name} = $count;
  }

# /proc/diskstats fields from kernel-doc/Documentation/iostats.txt
#
#Field  1 -- # of reads issued
#    This is the total number of reads completed successfully.
#Field  2 -- # of reads merged, field 6 -- # of writes merged
#    Reads and writes which are adjacent to each other may be merged for
#    efficiency.  Thus two 4K reads may become one 8K read before it is
#    ultimately handed to the disk, and so it will be counted (and queued)
#    as only one I/O.  This field lets you know how often this was done.
#Field  3 -- # of sectors read
#    This is the total number of sectors read successfully.
#Field  4 -- # of milliseconds spent reading
#    This is the total number of milliseconds spent by all reads (as
#    measured from __make_request() to end_that_request_last()).
#Field  5 -- # of writes completed
#    This is the total number of writes completed successfully.
#Field  7 -- # of sectors written
#    This is the total number of sectors written successfully.
#Field  8 -- # of milliseconds spent writing
#    This is the total number of milliseconds spent by all writes (as
#    measured from __make_request() to end_that_request_last()).
#Field  9 -- # of I/Os currently in progress
#    The only field that should go to zero. Incremented as requests are
#    given to appropriate request_queue_t and decremented as they finish.
#Field 10 -- # of milliseconds spent doing I/Os
#    This field is increases so long as field 9 is nonzero.
#Field 11 -- weighted # of milliseconds spent doing I/Os
#    This field is incremented at each I/O start, I/O completion, I/O
#    merge, or read of these stats by the number of I/Os in progress
#    (field 9) times the number of milliseconds spent doing I/O since the
#    last update of this field.  This can provide an easy measure of both
#    I/O completion time and the backlog that may be accumulating.

foreach my $d (@diskstat)
  {
    $d =~ s/^\s+//;
    my ($major,$minor,$disk,@tail) = split(/\s+/,$d);
    my $ntail = @tail;
    next unless $ntail == 11;
    next if ($tail[0] == 0);
    #print "[$disk]\n";
    next unless -e "/sys/block/$disk"; # report only real disks, omit partitions
    $disk =~ s/\W/\_/g;
    #print "$major $minor [$disk] $ntail @tail [$tail[0]]\n";
    $count{"dev_".$disk."_busy"} = $tail[10-1]/1000.0*100.0; # convert to seconds, then percents
    $count{"dev_".$disk."_read_sectors"} = $tail[3-1];
    $count{"dev_".$disk."_write_sectors"} = $tail[7-1];
    $count{"dev_".$disk."_read_rate_sectors"} = 1;
    $count{"dev_".$disk."_write_rate_sectors"} = 1;
    $count{"dev_".$disk."_io_rate_sectors"} = 1;
  }

#die "Here!";

my $tmpfile = $confDatafile;
if (-r $tmpfile)
  {
    open(FIN,$tmpfile);
    $xtime = int(<FIN>);
    my @fin = <FIN>;
    foreach my $in (@fin)
      {
        my ($name,$count) = split(/ +/,$in);
        $xcount{$name} = $count;
      }
    close FIN;

    die "Time goes backwards: $xtime -> $now" if ($xtime >= $now);
  }

my $dt = $now - $xtime;

#print "delta t is $dt sec\n";

die "Too little time ($dt sec) between calls to $0" if $dt < 10;

if (!$dryrun) {
  unlink $tmpfile;
  open(FOUT,">$tmpfile") || die "Cannot write to $tmpfile: $!";
  print FOUT "$now\n";
  foreach my $k (sort keys %count)
    {
      print FOUT "$k $count{$k}\n";
    }
  close FOUT;

  die "Too much time ($dt sec) between calls to $0" if $dt > 30*60;
}

my $boottime = $count{boot_time};
if ($boottime)
  {
    my $uptime = ($now - $boottime)/(60*60*24);
    $count{"uptime_days"} = $uptime;
    $xcount{"uptime_days"} = $uptime;
  }

foreach my $k (sort keys %count)
  {
    my $v1 = $count{$k};
    my $v2 = $xcount{$k};
    next unless $v1 && $v2;
    my $label = "vmstat_$k";
    my $val = $v1;
    my $rate;
    $rate = 1 if ($k =~ /ticks/);
    $rate = 1 if ($k =~ /pages_paged/);
    $rate = 1 if ($k =~ /pages_swapped/);
    $rate = 1 if ($k =~ /interrupts/);
    $rate = 1 if ($k =~ /forks/);
    $rate = 1 if ($k =~ /context_switches/);
    $rate = 1 if ($k =~ /sectors/);
    $rate = 1 if ($k =~ /busy/);
    if ($rate)
      {
        $val = ($v1 - $v2)/$dt;
        $label .= "_per_sec";
      }

    if ($k =~ /(.*)_read_rate_sectors/) {
      my $root = $1;
      my $r1 = $count{$root."_read_sectors"};
      my $r2 = $xcount{$root."_read_sectors"};
      my $dr = $r1-$r2;
      my $b1 = $count{$root."_busy"};
      my $b2 = $xcount{$root."_busy"};
      my $db = $b1-$b2;
      if ($b1 == 0 && $b2 == 0) {
	$val = -1;
      } elsif ($db == 0) {
	$val = 0;
      } else {
	$val = $dr/($db/100.0); # undo the percentage multiplication
      }

      $label = "vmstat_$k";
      #print "[$k] [$root] [$r1] [$r2] [$dr] [$b1] [$b2] [$db] val=$val\n" if $root =~ /sdi/;
    }

    if ($k =~ /(.*)_write_rate_sectors/) {
      my $root = $1;
      my $w1 = $count{$root."_write_sectors"};
      my $w2 = $xcount{$root."_write_sectors"};
      my $dw = $w1-$w2;
      my $b1 = $count{$root."_busy"};
      my $b2 = $xcount{$root."_busy"};
      my $db = $b1-$b2;
      if ($b1 == 0 && $b2 == 0) {
	$val = -1;
      } elsif ($db == 0) {
	$val = 0;
      } else {
	$val = $dw/($db/100.0); # undo the percentage multiplication
      }

      $label = "vmstat_$k";
      #print "[$k] [$root] [$w1] [$w2] [$dw] [$b1] [$b2] [$db] val=$val\n";
    }

    if ($k =~ /(.*)_io_rate_sectors/) {
      my $root = $1;
      my $r1 = $count{$root."_read_sectors"};
      my $r2 = $xcount{$root."_read_sectors"};
      my $dr = $r1-$r2;
      my $w1 = $count{$root."_write_sectors"};
      my $w2 = $xcount{$root."_write_sectors"};
      my $dw = $w1-$w2;
      my $b1 = $count{$root."_busy"};
      my $b2 = $xcount{$root."_busy"};
      my $db = $b1-$b2;
      if ($b1 == 0 && $b2 == 0) {
	$val = -1;
      } elsif ($db == 0) {
	$val = 0;
      } else {
	$val = ($dr+$dw)/($db/100.0); # undo the percentage multiplication
      }

      $label = "vmstat_$k";
      #print "[$k] [$root] [$w1] [$w2] [$dw] [$b1] [$b2] [$db] val=$val\n";
    }

    next if $val < 0;
    my $keep = 0;
    $keep = 1 if $label =~ /read_sectors/;
    $keep = 1 if $label =~ /write_sectors/;
    $keep = 1 if $label =~ /rate_sectors/;
    $keep = 1 if $label =~ /busy/;
    $keep = 1 if $label =~ /pages_paged/;
    $keep = 1 if $label =~ /pages_swapped/;
    $keep = 1 if $label =~ /uptime_days/;
    next unless $keep;
    my $exec = "$confGmetric $gmetricDest -tfloat -n\"$label\" -v\"$val\"";
    print "$exec\n" if $verbose;
    system $exec unless $dryrun;
  }

exit 0;
#end file
